// Author: A.Voss@FH-Aachen.de
//
// memory classes
// tags: extern, auto, static, register, mutable

#include <iostream>
#include <cstdlib>
using namespace std;

#include "../../include/tools.h"

int main()
{
    // Speicherklasse auto
    // Bsp. auto int n;
    // old-C-style: Variablen vom Typ auto werden beim Aufruf einer Funktion 
    // erzeugt und danach wieder zerstoert ... das ist das default-Verhalten
    // fuer lokale Variablen, so dass auto in der ursprünglichen Bedeutung
    // obsolet wurde und in C++ eine neue Bedeutung hat, naemlich die automatische
    // Bestimmung des Variablentyps. Ihr Geltungsbereich war die lokale Funktion.

    // Speicherklasse extern
    // Bsp. extern int n;
    // Objekte, die ausserhalb von Funktionen deklariert werden, sind von der
    // Speicherklasse extern und sogar in anderen Modulen sichtbar.
    // Obiges Bsp. bedeutet, es gibt irgendwo eine Variable namens n vom Typ int, 
    // aber hier wird die Variable nur deklariert, nicht definiert. Wichtig ist 
    // auch, dass es nicht mehrere davon gibt -> siehe Link-Beispiele.

    // Speicherklasse static
    // Bsp. static int n;
    // Objekte der Speicherklasse static koennen sowohl interne als auch externe
    // sein. Innerhalb von Funktionen behalten sie ihre Werte zwischen den 
    // Funktionsaufrufen -> siehe static-Beispiele.
    // Funktionen, die als static dekl. sind, sind nur in dem Modul sichtbar.
    // Zeichenketten innerhalb von Funktionen sind static. Static Objekte ausserhalb
    // von Funktionen sind nur in dem Modul bzw. der Quellcodedatei bekannt.
    // Statische Variablen in Klassen gibt es nur einmal. Statische Funktionen
    // in Klassen koennen ohne Objekte und nur ueber die Klasse aufgerufen
    // werden.

    // Speicherklasse register
    // Bsp. register int n;
    // Die Angabe der Speicherklasse register legt dem Compiler nahe, die Variable
    // oder den Parameter in einem Prozessorregister abzulegen. Das kann aber ignoriert
    // werden oder schlicht unmoeglich sein. Typischerweise kann der Compiler
    // von alleine besser optimieren... Ansonsten wie auto.
 
    // Speicherklasse mutable
    // mutable bei Klassenmembern gibt an, dass member veraendert werden koennen,
    // obwohl das Objekt const ist.
    
        
    register int n=12;

    tools_log(); cout << "n=" << n << endl;
    
    class C
    {
    public:
        mutable int n;
        C() : n(12) { }
        int get_n() const { return n++; }
    };

    C c;
    tools_log(); cout << "c.n=" << c.get_n() << endl;
    tools_log(); cout << "c.n=" << c.get_n() << endl;
    
    return EXIT_SUCCESS;
}

