// Author: A.Voss@FH-Aachen.de
//
// bits and operators, fast eval
// tags: bits

#include <iostream>
#include <memory>
#include <cstdlib>
#include <vector>
using namespace std;

#include "../../include/tools.h"

#pragma GCC diagnostic ignored "-Wdeprecated-declarations"

// http://ootips.org/yonat/4dev/smart-pointers.html

class C {
public:
    int n;
    
    C(int N) : n(N) { 
        tools_log(); cout << "ctor " << n << endl; 
    }
    ~C() { 
        tools_log(); cout << "dtor " << n << endl; 
    }
    
    friend ostream & operator<<(ostream & os, const C & c) { 
        os << c.n; return os; 
    }
};

// some factory function
C * createC(int n) { return new C(n); }

int main()
{
    tools_log(); cout << "some operations..." << endl;

    {
        C c(1);
        tools_log(); cout << "c=" << c << endl;
    }

    // we are responsible for destroying dynamic objects!
    {
        C* p = createC(2);
        tools_log(); cout << "*p=" << *p << endl;
        delete p;
    }

    // one solution for c++ versions < c++11
    {
        auto_ptr<C> p(createC(3));
        tools_log(); cout << "*p=" << *p << endl;
        // => the pointer is responsible for the object and deletes it
        //    at its end
    }
    // note: - auto_ptrs are deprecated in C++11, there shared_ptr, unique_ptr
    //         or weak_ptr should be used
    //       - they store a single pointer to an object and destroy it 
    //         automatically at end of scope
    //       - they implement "strict ownership", i.e. to copy an auto_ptr
    //         means tranfering ownership/responsibility to the copy
    //       - auto_ptrs delete objects via "delete", so auto_ptrs are designed for
    //         pointers comming from "new" operations
    //       - they should not be used with stl containers involving element copies

    {
        auto_ptr<C> p1(createC(4));    // p1 owns C-object
        auto_ptr<C> p2;
    
        // we cannot deref. p2!
        tools_log(); cout << "p1=" << (*p1) << " &p1:" << p1.get() << endl;
        tools_log(); cout << "p2=" << "-"   << " &p2:" << p2.get() << endl;
    
        // now we transfer ownership from p1 to p2
        // hence we cannot deref. p1 anymore
        p2 = p1;
        tools_log(); cout << "p1=" << "-"   << " &p1:" << p1.get() << endl;
        tools_log(); cout << "p2=" << (*p2) << " &p1:" << p2.get() << endl;
        
        // modify object
        p2->n = 5;
        tools_log(); cout << "p2=" << (*p2) << " &p1:" << p2.get() << endl;

        // it is also possible to declare a pointer to a const object
        // which cannot be modified
        // auto_ptr<const C> p3(p2);

        vector< auto_ptr<C> > v;
        v.push_back( auto_ptr<C>(new C(6)) );
        
        auto_ptr<C> p3 = v[0];     // here v[0] looses ownership
        tools_log(); cout << "v[0].get:" << v[0].get() << endl;
        tools_log(); cout << "p3=" << (*p3) << " &p3:" << p3.get() << endl;
        
        //auto_ptr<C> p4 ( new C[1] );  // whats the problem here?
    }
    
    return EXIT_SUCCESS;
}
