// Author: A.Voss@FH-Aachen.de
//
// bits and operators, fast eval
// tags: bits

#include <iostream>
#include <memory>
#include <cstdlib>
#include <vector>
using namespace std;

#include "../../include/tools.h"

// http://ootips.org/yonat/4dev/smart-pointers.html

class C {
public:
    int n;
    
    C(int N=0) : n(N) { 
        tools_log(); cout << "ctor " << n << endl; 
    }
    ~C() { 
        tools_log(); cout << "dtor " << n << endl; 
    }
    
    friend ostream & operator<<(ostream & os, const C & c) { 
        os << c.n; return os; 
    }
};

// some factory function
C * createC(int n) { return new C(n); }

int main()
{
    tools_log(); cout << "some operations..." << endl;

    {
        unique_ptr<C> p1(createC(4));    // p1 owns C-object
        unique_ptr<C> p2;
    
        // we cannot deref. p2!
        tools_log(); cout << "p1=" << (*p1) << " &p1:" << p1.get() << endl;
        tools_log(); cout << "p2=" << "-"   << " &p2:" << p2.get() << endl;
    
        // now we transfer ownership from p1 to p2
        // hence we cannot deref. p1, and the following is 
        // not valid anymore (as the assignment-op is deleted):
        // p2 = p1;
        p2 = move(p1);
        tools_log(); cout << "p1=" << "-"   << " &p1:" << p1.get() << endl;
        tools_log(); cout << "p2=" << (*p2) << " &p1:" << p2.get() << endl;
        
        // modify object
        p2->n = 5;
        tools_log(); cout << "p2=" << (*p2) << " &p1:" << p2.get() << endl;

        vector< unique_ptr<C> > v;
        v.push_back( auto_ptr<C>(new C(6)) );
        
        unique_ptr<C> p3 = move(v[0]);     
        tools_log(); cout << "v[0].get:" << v[0].get() << endl;
        tools_log(); cout << "p3=" << (*p3) << " &p3:" << p3.get() << endl;

        // works for arrays
        unique_ptr<C[]> p4 ( new C[3] );
    }
    
    return EXIT_SUCCESS;
}
